#parses response from gateway

imports = Imports(
	{
		"StartParse": "discum.gateway.start.parse",
		"GuildParse": "discum.gateway.guild.parse",
		"UserParse": "discum.gateway.user.parse",
		"MessageParse": "discum.gateway.messages.parse",
		"ChannelParse": "discum.gateway.channels.parse",
	}
)

import copy

#function names are just lowercase types, so for type GUILD_MEMBER_LIST_UPDATE, the function is guild_member_list_update
class Parse(object):
	__slots__ = ['response']
	def __init__(self, response):
		self.response = copy.deepcopy(response)

	def auto(self): #auto parse, does not allow for custom inputs
		if hasattr(self, str(self.response['t']).lower()):
			return getattr(self, str(self.response['t']).lower())()
		return self.response['d'] #just return the value of d if there's no parse function for it yet

	def ready(self):
		return imports.StartParse().ready(self.response)

	def ready_supplemental(self):
		return imports.StartParse().ready_supplemental(self.response)

	def guild_member_list_update(self):
		return imports.GuildParse().guild_member_list_update(self.response)

	def guild_create(self, my_user_id="0"): #personal user id needed to update personal roles for that guild
		return imports.GuildParse().guild_create(self.response, my_user_id)

	def guild_members_chunk(self):
		return imports.GuildParse().guild_members_chunk(self.response)

	def message_create(self):
		return imports.MessageParse().message_create(self.response)

	def sessions_replace(self, session_id="0"):
		return imports.UserParse().sessions_replace(self.response, session_id)

	def channel_create(self):
		return imports.ChannelParse().channel_create(self.response)

	def channel_delete(self):
		return imports.ChannelParse().channel_delete(self.response)



from curl_cffi.requests import Response


__all__ = [
    "TwitterException",
    "HTTPException",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "RateLimited",
    "TwitterServerError",
]


class TwitterException(Exception):
    pass


class HTTPException(TwitterException):
    """Exception raised when an HTTP request fails.
    """

    def __init__(self, response: Response, response_json: dict = None):
        self.response = response
        self.api_errors: list[dict[str, int | str]] = []
        self.api_codes: list[int] = []
        self.api_messages: list[str] = []

        if response_json is None:
            super().__init__(f"{response.status_code}")
            return

        errors = response_json.get("errors", [])

        if "error" in response_json:
            errors.append(response_json["error"])
        else:
            errors.append(response_json)

        error_text = ""

        for error in errors:
            self.api_errors.append(error)

            if isinstance(error, str):
                self.api_messages.append(error)
                error_text += '\n' + error
                continue

            if "code" in error:
                self.api_codes.append(error["code"])
            if "message" in error:
                self.api_messages.append(error["message"])

            if "code" in error and "message" in error:
                error_text += f"\n{error['code']} - {error['message']}"
            elif "message" in error:
                error_text += '\n' + error["message"]

        if not error_text and "detail" in response_json:
            self.api_messages.append(response_json["detail"])
            error_text = '\n' + response_json["detail"]

        super().__init__(f"{response.status_code} {error_text}")


class BadRequest(HTTPException):
    """Exception raised for a 400 HTTP status code.
    """
    pass


class Unauthorized(HTTPException):
    """Exception raised for a 401 HTTP status code.
    """
    pass


class Forbidden(HTTPException):
    """Exception raised for a 403 HTTP status code.
    """
    pass


class NotFound(HTTPException):
    """Exception raised for a 404 HTTP status code.
    """
    pass


class RateLimited(HTTPException):
    """Exception raised for a 429 HTTP status code.
    """
    pass


class TwitterServerError(HTTPException):
    """Exception raised for a 5xx HTTP status code.
    """
    pass



#request

class MediaRequest(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def call(self, channelID, guildID=None, mute=False, deaf=False, video=False):
		self.gatewayobj.send(
		    {
		        "op": self.gatewayobj.OPCODE.VOICE_STATE_UPDATE,
		        "d": {
		            "guild_id": guildID,
		            "channel_id": channelID,
		            "self_mute": mute,
		            "self_deaf": deaf,
		            "self_video": video,
		        },
		    }
		)

	def endCall(self):
		self.gatewayobj.send(
		    {
		        "op": self.gatewayobj.OPCODE.VOICE_STATE_UPDATE,
		        "d": {
		            "guild_id": None,
		            "channel_id": None,
		            "self_mute": False,
		            "self_deaf": False,
		            "self_video": False,
		        },
		    }
		)




from ..types import Types

#parse
class ChannelParse(object): #can either be a guild channel or a DM or something else
	@staticmethod
	def channel_create(response):
		channelData = dict(response["d"])
		channelData["type"] = Types.CHANNEL[response["d"]["type"]]
		if channelData["type"] in ("dm", "group_dm"): #private_channel
			if "recipient_ids" not in channelData and "recipients" in channelData: #should be true, running this check just in case
				channelData["recipient_ids"] = [i["id"] for i in channelData["recipients"]]
		return channelData

	@staticmethod
	def channel_delete(response):
		channelData = dict(response["d"])
		channelData["type"] = Types.CHANNEL[response["d"]["type"]]
		if channelData["type"] in ("dm", "group_dm"): #private_channel
			if "recipient_ids" not in channelData and "recipients" in channelData:
				channelData["recipient_ids"] = [i["id"] for i in channelData["recipients"]]
		return channelData


from .file import (
    copy_file,
    load_lines,
    load_json,
    load_toml,
    write_lines,
    write_json,
    to_json,
)
from .asyncio import (
    set_windows_selector_event_loop_policy,
    curry_async,
    bounded_gather,
)


__all__ = [
    "copy_file",
    "load_lines",
    "load_json",
    "load_toml",
    "write_lines",
    "write_json",
    "to_json",
    "set_windows_selector_event_loop_policy",
    "curry_async",
    "bounded_gather",
]



"""
discum: message types, channel types, etc
 only for objects with the "type" key
"""


class Types:
    MESSAGE = {
        0: "default",
        1: "recipient_added",
        2: "recipient_removed",
        3: "call",
        4: "channel_name_changed",
        5: "channel_icon_changed",
        6: "channel_message_pinned",
        7: "guild_member_joined",
        8: "user_premium_guild_subscription",
        9: "user_premium_guild_subscription_tier_1",
        10: "user_premium_guild_subscription_tier_2",
        11: "user_premium_guild_subscription_tier_3",
        12: "channel_follow_added",
        14: "guild_discovery_disqualified",
        15: "guild_discovery_requalified",
        16: "guild_discovery_grace_period_initial_warning",
        17: "guild_discovery_grace_period_final_warning",
        18: "thread_created",
        19: "reply",
        20: "application_command",
        21: "thread_starter_message",
        22: "guild_invite_reminder",
        23: "context_menu_command"
    }

    CHANNEL = {
        0: "guild_text",
        1: "dm",
        2: "guild_voice",
        3: "group_dm",
        4: "guild_category",
        5: "guild_news",
        6: "guild_store",
        10: "guild_news_thread",
        11: "guild_public_thread",
        12: "guild_private_thread",
        13: "guild_stage_voice"
    }

    RELATIONSHIP = {
        1: "friend",
        2: "blocked",
        3: "pending_incoming",
        4: "pending_outgoing"
    }



#request

class DmRequest(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def DMchannel(self, channel_id):
		self.gatewayobj.send({"op":self.gatewayobj.OPCODE.DM_UPDATE,"d":{"channel_id":channel_id}})


from curl_cffi import requests


class BaseAsyncSession(requests.AsyncSession):
    DEFAULT_HEADERS = {
        "accept": "*/*",
        "accept-language": "en-US,en",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Safari/537.36",
        "sec-ch-ua": '"Not_A Brand";v="8", "Chromium";v="110", "Google Chrome";v="110"',
        "sec-ch-ua-platform": '"Windows"',
        "sec-ch-ua-mobile": "?0",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
    }
    DEFAULT_IMPERSONATE = requests.BrowserType.chrome110
    """
    Базовый асинхронная сессия:
        - Принимает прокси в стандартном URL-формате вместо словаря.
        - По умолчанию устанавливает версию браузера chrome110.
        - По умолчанию устанавливает user-agent под версию браузера chrome110.
    """

    def __init__(
            self,
            proxy: str = None,
            **session_kwargs,
    ):
        proxies = {"http": proxy, "https": proxy}
        headers = session_kwargs["headers"] = session_kwargs.get("headers") or {}
        headers.update(self.DEFAULT_HEADERS)
        session_kwargs["impersonate"] = session_kwargs.get("impersonate") or self.DEFAULT_IMPERSONATE
        super().__init__(proxies=proxies, **session_kwargs)

    @property
    def user_agent(self) -> str:
        return self.headers["user-agent"]



async def get_oauth_auth_tokens(self) -> tuple[str | None, str | None, str | None, ClientResponse]:
    while True:
        # noinspection PyProtectedMember
        headers: dict = self.twitter_client._headers

        if headers.get('content-type'):
            del headers['content-type']
        headers[
            'accept'] = ('text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,'
                         '*/*;q=0.8,application/signed-exchange;v=b3;q=0.7')

        if not self.twitter_client.ct0:
            # noinspection PyProtectedMember
            self.twitter_client.set_ct0(await self.twitter_client._request_ct0())

        while True:
            try:
                r = await self.twitter_client.request(url='https://memefarm-api.memecoin.org/user/twitter-auth',
                                                      method='get',
                                                      params={
                                                          'callback': 'https://www.memecoin.org/farming'
                                                      },
                                                      headers=headers)
            except better_automation.twitter.errors.BadRequest as error:
                logger.error(f'{self.account_token} | BadRequest: {error}, пробую еще раз')
            else:
                break
        if BeautifulSoup(await r[0].text(), 'lxml').find('iframe', {
            'id': 'arkose_iframe'
        }):
            logger.info(f'{self.account_token} | Обнаружена капча на аккаунте, пробую решить')
            if not await SolveCaptcha(auth_token=self.twitter_client.auth_token,
                                      ct0=self.twitter_client.ct0).solve_captcha(
                proxy=Proxy.from_str(proxy=self.account_proxy).as_url if self.account_proxy else None):
                raise exceptions.WrongCaptcha()
            continue

        if 'https://www.memecoin.org/farming?oauth_token=' in (await r[0].text()):
            return 'https://www.memecoin.org/farming?oauth_token=' + \
                   (await r[0].text()).split('https://www.memecoin.org/farming?oauth_token=')[-1].split('"')[
                       0].replace('&amp;', '&'), None, None, r[0]

        auth_token_html = BeautifulSoup(await r[0].text(), 'lxml').find('input', {
            'name': 'authenticity_token'
        })
        oauth_token_html = BeautifulSoup(await r[0].text(), 'lxml').find('input', {
            'name': 'oauth_token'
        })

        if not auth_token_html or not oauth_token_html:
            logger.error(f'{self.account_token} | Не удалось обнаружить Auth/OAuth Token на странице, '
                         f'статус: {r[0].status}')
            return None, None, None, r[0]

        auth_token: str = auth_token_html.get('value', '')
        oauth_token: str = oauth_token_html.get('value', '')
        return None, auth_token, oauth_token, r[0]


async def get_auth_location(self,
                            oauth_token: str,
                            auth_token: str) -> tuple[str | bool, str]:
    while True:
        if not self.twitter_client.ct0:
            # noinspection PyProtectedMember
            self.twitter_client.set_ct0(await self.twitter_client._request_ct0())
        # noinspection PyProtectedMember
        r = await self.twitter_client.request(url='https://api.twitter.com/oauth/authorize',
                                              method='post',
                                              data={
                                                  'authenticity_token': auth_token,
                                                  'redirect_after_login': f'https://api.twitter.com/oauth'
                                                                          f'/authorize?oauth_token={oauth_token}',
                                                  'oauth_token': oauth_token
                                              },
                                              headers={
                                                  **self.twitter_client._headers,
                                                  'content-type': 'application/x-www-form-urlencoded'
                                              })
        if 'This account is suspended' in await r[0].text():
            raise AccountSuspended(self.account_token)

        if 'https://www.memecoin.org/farming?oauth_token=' in await r[0].text():
            location: str = 'https://www.memecoin.org/farming?oauth_token=' + \
                            (await r[0].text()).split('https://www.memecoin.org/farming?oauth_token=')[-1].split(
                                '"')[0].replace('&amp;', '&')
            return location, await r[0].text()
        return False, await r[0].text()


async def make_old_auth(self,
                        oauth_token: str,
                        oauth_verifier: str) -> tuple[int, str, Response | ClientResponse]:
    while True:
        r = self.meme_client.post(url='https://memefarm-api.memecoin.org/user/twitter-auth1',
                                  json={
                                      'oauth_token': oauth_token,
                                      'oauth_verifier': oauth_verifier
                                  })

        if r.json().get('error', '') == 'account_too_new':
            return 1, '', r

        elif r.json().get('error', '') == 'Unknown Error':
            if await self.check_captcha():
                logger.info(
                    f'{self.account_token} | Обнаружена капча на аккаунте, пробую решить')

                if not await SolveCaptcha(auth_token=self.twitter_client.auth_token,
                                          ct0=self.twitter_client.ct0).solve_captcha(
                    proxy=Proxy.from_str(
                        proxy=self.account_proxy).as_url if self.account_proxy else None):
                    raise exceptions.WrongCaptcha()
                continue

            if r.json().get('status', 0) and r.json()['status'] == 429:
                logger.error(f'{self.account_token} | Неизвестный ответ при авторизации MEME: {r.text}')
                await asyncio.sleep(delay=5)
                continue

            return 2, '', r

        elif r.json().get('error', '') in ['unauthorized',
                                           'Unauthorized']:
            # noinspection PyTypeHints
            r.status: int = r.status_code
            # noinspection PyTypeHints
            r.reason: str = ''
            raise better_automation.twitter.errors.Unauthorized(r)

        elif r.json().get('accessToken', ''):
            return 0, r.json()['accessToken'], r

        return 2, '', r


async def request_access_token_old(self) -> tuple[int | None, str | None, ClientResponse | None]:
    location, auth_token, oauth_token, r = await self.get_oauth_auth_tokens()

    if not location:
        if not auth_token \
                or not oauth_token:
            logger.error(
                f'{self.account_token} | Ошибка при получении OAuth / Auth Token, '
                f'статус: {r.status}')
            return None, None, r
        location, response_text = await self.get_auth_location(oauth_token=oauth_token,
                                                               auth_token=auth_token)
        if not location:
            logger.error(
                f'{self.account_token} | Ошибка при авторизации через Twitter, '
                f'статус: {r.status}')
            return None, None, r

    if parse_qs(urlparse(location).query).get('redirect_after_login') \
            or not parse_qs(urlparse(location).query).get('oauth_token') \
            or not parse_qs(urlparse(location).query).get('oauth_verifier'):
        logger.error(
            f'{self.account_token} | Не удалось обнаружить OAuth Token / OAuth Verifier в '
            f'ссылке: {location}')
        return None, None, r

    oauth_token: str = parse_qs(urlparse(location).query)['oauth_token'][0]
    oauth_verifier: str = parse_qs(urlparse(location).query)['oauth_verifier'][0]

    return await self.make_old_auth(oauth_token=oauth_token,
                                    oauth_verifier=oauth_verifier)


##################################################################

def login(self):

    response = self.session.get(f"https://api-invite.kuscription.com/v1/info?account={self.address}")

    code = response.json()['data']['code']
    # print(code)

    if response.json()['data']['socials'] == []:
        response = self.session.get("https://api-invite.kuscription.com/v1/login_twitter")
        authUrl = response.json()['data']

        self.session.cookies.update({'auth_token': self.auth_token, 'ct0': self.csrf})
        response = self.session.get(authUrl)

        soup = BeautifulSoup(response.text, 'html.parser')
        authenticity_token = soup.find('input', attrs={'name': 'authenticity_token'}).get('value')

        payload = {'authenticity_token': authenticity_token,
                   'redirect_after_login': authUrl,
                   'oauth_token': authUrl.split("oauth_token=")[-1]}

        response = self.session.post(f'https://api.twitter.com/oauth/authorize', data=payload,
                                     headers={'content-type': 'application/x-www-form-urlencoded'})
        soup = BeautifulSoup(response.text, 'html.parser')
        link = soup.find('a', class_='maintain-context').get('href')
        # print(link)

        response = self.session.get(link)

        verifier = link.split("oauth_verifier=")[-1]
        response = self.session.post("https://api-invite.kuscription.com/v1/auth_twitter",
                                     json={"account": self.address,
                                           "token": authUrl.split("oauth_token=")[-1],
                                           "verifier": verifier},
                                     headers={'content-type': 'application/json'})

        # if response.json()['msg'] != "SUCCESS":
        #     raise Exception("Error 1")

        # print(response.text)

    return code


#points to commands that help request info/actions using the gateway
#note, no need for importing GuildParse because resp is a Resp object (resp.parsed... does the trick)
#also, no need for importing GuildRequest because gatewayobj has that (self.gatewayobj.request... does the trick)

import time
import copy
import re
from ...utils.permissions import PERMS, Permissions
from ...logger import *

class GuildCombo(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	#fetchMembers helper function
	def reformat_member(self, memberdata, keep=[]): #memberdata comes in as a dict and leaves as a tuple (userID, memberdatareformatted). This is done to easily prevent duplicates in the member list when fetching.
		allProperties = ['pending', 'deaf', 'hoisted_role', 'presence', 'joined_at', 'public_flags', 'username', 'avatar', 'discriminator', 'premium_since', 'roles', 'is_pending', 'mute', 'nick', 'bot', 'communication_disabled_until']
		if keep == None:
			remove = allProperties
		elif keep == "all":
			remove = []
		elif isinstance(keep, list) or isinstance(keep, tuple):
			remove = list(set(allProperties) - set(keep))
		elif isinstance(keep, str):
			remove = [i for i in allProperties if i!=keep]
		memberproperties = copy.deepcopy(memberdata['member']) if 'member' in memberdata else copy.deepcopy(memberdata)
		userdata = memberproperties.pop('user', {})
		userID = userdata.pop('id', {})
		memberproperties.update(userdata)
		#filtering/removing
		for r in remove:
			if r in memberproperties:
				del memberproperties[r]
		return userID, memberproperties

	#fetchMembers helper function
	def rangeCorrector(self, ranges): #just adds [0,99] at the beginning
		if [0,99] not in ranges:
			ranges.insert(0, [0,99])
		return ranges

	#fetchMembers helper function
	def getIndex(self, guild_id):
		return self.gatewayobj.memberFetchingStatus[guild_id][1]

	#fetchMembers helper function
	def getRanges(self, index, multiplier, memberCount):
		initialNum = int(index*multiplier)
		rangesList = [[initialNum, initialNum+99]]
		if memberCount > initialNum+99:
			rangesList.append([initialNum+100, initialNum+199])
		return self.rangeCorrector(rangesList)

	#fetchMembers helper function
	def updateCurrent(self, guild_id):
		if not self.gatewayobj.finishedMemberFetching(guild_id): #yep still gotta check for this
			self.gatewayobj.memberFetchingStatus[guild_id][1] = self.gatewayobj.memberFetchingStatus[guild_id][0]+1

	#fetchMembers helper function
	def updatePrevious(self, guild_id):
		if not self.gatewayobj.finishedMemberFetching(guild_id):
			self.gatewayobj.memberFetchingStatus[guild_id][0] = self.gatewayobj.memberFetchingStatus[guild_id][1]

	#todo: make channel_id optional (make a helper method to find the "optimal" channel). Also...maybe rewrite fetchMembers to simply code a bit??
	def fetchMembers(self, resp, guild_id, channel_id, method, keep, considerUpdates, startIndex, stopIndex, reset, wait): #process is a little simpler than it looks. Keep in mind that there's no actual api endpoint to get members so this is a bit hacky. However, the method used below mimics how the official client loads the member list.
		if self.gatewayobj.READY:
			if self.gatewayobj.memberFetchingStatus.get(guild_id) == None: #request for lazy request
				self.gatewayobj.memberFetchingStatus[guild_id] = [startIndex, startIndex] #format is [previous index, current index]. This format is useful for the wait parameter.
				if not self.gatewayobj.session.guild(guild_id).hasMembers and reset:
					self.gatewayobj.session.guild(guild_id).resetMembers() #reset
				if len(self.gatewayobj.memberFetchingStatus["first"]) == 0:
					self.gatewayobj.memberFetchingStatus["first"] = [guild_id]
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: [[0,99]]}, typing=True, threads=False, activities=True, members=[])
				else:
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: [[0,99]]}, typing=True, activities=True)
			if self.gatewayobj.memberFetchingStatus.get(guild_id) != None and not self.gatewayobj.finishedMemberFetching(guild_id): #proceed with lazy requests
				index = self.getIndex(guild_id) #index always has the current value
				endFetching = False
				#find multiplier (this dictates the way the member list requested for)
				if method == "overlap": multiplier = 100
				elif method == "no overlap": multiplier = 200
				elif isinstance(method, int): multiplier = method
				elif isinstance(method, list) or isinstance(method, tuple): 
					if index<len(method):
						multiplier = method[index]
					else:
						endFetching = True #ends fetching right after resp parsed
				ranges = self.getRanges(index, multiplier, self.gatewayobj.session.guild(guild_id).memberCount) if not endFetching else [[0],[0]]
				#0th lazy request (separated from the rest because this happens "first")
				if index == startIndex and not self.gatewayobj.session.guild(guild_id).unavailable:
					self.updateCurrent(guild_id) #current = previous+1
					if wait!=None: time.sleep(wait)
					self.gatewayobj.request.lazyGuild(guild_id, {channel_id: ranges})
				if resp.event.guild_member_list:
					parsed = resp.parsed.guild_member_list_update()
					if parsed['guild_id'] == guild_id and ('SYNC' in parsed['types'] or 'UPDATE' in parsed['types']):
						endFetching = False
						for ind,i in enumerate(parsed['types']):
							if i == 'SYNC':
								if len(parsed['updates'][ind]) == 0 and parsed['locations'][ind] in ranges[1:]: #checks if theres nothing in the SYNC data
									endFetching = True
									break
								for item in parsed['updates'][ind]:
									if 'member' in item:
										member_id, member_properties = self.reformat_member(item, keep=keep)
										self.gatewayobj.session.guild(guild_id).updateOneMember(member_id, member_properties)
										Logger.log('[gateway] [fetchMembers] <SYNC> updated member '+member_id, None, self.gatewayobj.log)
								if not self.gatewayobj.finishedMemberFetching(guild_id) and (index-self.gatewayobj.memberFetchingStatus[guild_id][0])==1:
									if wait!=None: time.sleep(wait)
									self.updatePrevious(guild_id) #previous = current
							elif i == 'UPDATE' and considerUpdates: #this really only becomes useful for large guilds (because fetching members can take a quite some time for those guilds)
								for key in parsed['updates'][ind]:
									if key == 'member':
										member_id, member_properties = self.reformat_member(parsed['updates'][ind][key], keep=keep)
										self.gatewayobj.session.guild(guild_id).updateOneMember(member_id, member_properties)
										Logger.log('[gateway] [fetchMembers] <UPDATE> updated member '+member_id, None, self.gatewayobj.log)
							elif i == 'INVALIDATE':
								if parsed['locations'][ind] in ranges or parsed['member_count'] == 0:
									endFetching = True
									break
						numFetched = len(self.gatewayobj.session.guild(guild_id).members)
						roundedUpFetched = numFetched-(numFetched%-100) #https://stackoverflow.com/a/14092788/14776493
						if ranges==[[0],[0]] or index>=stopIndex or roundedUpFetched>=self.gatewayobj.session.guild(guild_id).memberCount or endFetching or ranges[1][0]+100>self.gatewayobj.session.guild(guild_id).memberCount: #putting whats most likely to happen first
							self.gatewayobj.memberFetchingStatus[guild_id] = "done"
							self.gatewayobj.removeCommand(
							    {
							        "function": self.fetchMembers,
							        "params": {
							            "guild_id": guild_id,
							            "channel_id": channel_id,
							            "method": method,
							            "keep": keep,
							            "considerUpdates": considerUpdates,
							            "startIndex": startIndex,
							            "stopIndex": stopIndex,
							            "reset": reset,
							            "wait": wait
							        },
							    }
							) #it's alright if you get a "not found in _after_message_hooks" error log. That's not an error for this situation.
						elif not self.gatewayobj.finishedMemberFetching(guild_id) and index==self.gatewayobj.memberFetchingStatus[guild_id][0]:
							self.updateCurrent(guild_id) #current = previous + 1
							self.gatewayobj.request.lazyGuild(guild_id, {channel_id: ranges})


	#helper method for subscribeToGuildEvents
	def findVisibleChannels(self, guildID, types, findFirst):
		channelIDs = []
		if types == "all":
			types = ['guild_text', 'dm', 'guild_voice', 'group_dm', 'guild_category', 'guild_news', 'guild_store', 'guild_news_thread', 'guild_public_thread', 'guild_private_thread', 'guild_stage_voice']
		s = self.gatewayobj.session
		channels = s.guild(guildID).channels
		for channel in channels.values():
			if channel['type'] in types:
				permissions = Permissions.calculatePermissions(s.user['id'], guildID, s.guild(guildID).owner, s.guild(guildID).roles, s.guild(guildID).me['roles'], channel["permission_overwrites"])
				if Permissions.checkPermissions(permissions, PERMS.VIEW_CHANNEL):
					if findFirst:
						return [channel['id']]
					else:
						channelIDs.append(channel['id'])
		return channelIDs

	def subscribeToGuildEvents(self, onlyLarge, wait):
		if self.gatewayobj.READY:
			s = self.gatewayobj.session
			guildIDs = s.guildIDs
			first = {"channel_ranges":{}, "typing":True, "threads":True, "activities":True, "members":[], "thread_member_lists":[]}
			rest = {"channel_ranges":{}, "typing":True, "activities":True, "threads":True}
			for guildID in guildIDs:
				#skip if needed (onlyLarge checking)
				if onlyLarge and not (s.guild(guildID).unavailable or s.guild(guildID).large):
					continue
				#op 14 field construction
				op14fields = {"guild_id":guildID}
				if guildID == guildIDs[0]:
					op14fields.update(first)
					if not s.guild(guildID).unavailable:
						findChannel = self.findVisibleChannels(guildID, types="all", findFirst=True)
						if findChannel:
							op14fields["channel_ranges"] = {findChannel[0]: [[0,99]]}
				else:
					op14fields.update(rest)
					if not s.guild(guildID).unavailable:
						findChannel = self.findVisibleChannels(guildID, types="all", findFirst=True)
						if findChannel:
							op14fields["channel_ranges"] = {findChannel[0]: [[0,99]]}
				#sending the request
				if wait: time.sleep(wait)
				self.gatewayobj.memberFetchingStatus["first"].append(guildID)
				self.gatewayobj.request.lazyGuild(**op14fields)

	#helper for searchGuildMembers
	def handleGuildMemberSearches(self, resp, guildIDs, saveAsQuery, isQueryOverridden, userIDs, keep): #hm what happens if all userIDs are found? well good news: "not_found" value is just []
		if resp.event.guild_members_chunk:
			chunk = resp.parsed.auto()
			gID = chunk["guild_id"]
			match = False
			if gID in guildIDs:
				if userIDs and "not_found" in chunk:
					match = True
					for member in chunk["members"]:
						member_id, member_properties = self.reformat_member(member, keep=keep)
						self.gatewayobj.guildMemberSearches[gID]["ids"].add(member_id)
						self.gatewayobj.session.guild(gID).updateOneMember(member_id, member_properties)
				elif not userIDs:
					if isQueryOverridden:
						match = True #no checks
						for member in chunk["members"]:
							member_id, member_properties = self.reformat_member(member, keep=keep)
							self.gatewayobj.guildMemberSearches[gID]["queries"][saveAsQuery].add(member_id)
							self.gatewayobj.session.guild(gID).updateOneMember(member_id, member_properties)
					else: #check results
						if all([(re.sub(' +', ' ', k["user"]["username"].lower()).startswith(saveAsQuery) or re.sub(' +', ' ', k["nick"].lower()).startswith(saveAsQuery)) if k.get('nick') else re.sub(' +', ' ', k["user"]["username"].lower()).startswith(saveAsQuery) for k in chunk["members"]]): #search user/nick, ignore case, replace consecutive spaces with 1 space
							match = True
							for member in chunk["members"]:
								member_id, member_properties = self.reformat_member(member, keep=keep)
								self.gatewayobj.guildMemberSearches[gID]["queries"][saveAsQuery].add(member_id)
								self.gatewayobj.session.guild(gID).updateOneMember(member_id, member_properties)
				if chunk["chunk_index"] == chunk["chunk_count"]-1 and gID==guildIDs[-1]: #if at end
					if match:
						self.gatewayobj.removeCommand(
							{
								"function": self.handleGuildMemberSearches,
								"params": {
									"guildIDs": guildIDs,
									"saveAsQuery": saveAsQuery,
									"isQueryOverridden": isQueryOverridden,
									"userIDs": userIDs, 
									"keep": keep
								},
							}
						)

	def searchGuildMembers(self, guildIDs, query, saveAsQueryOverride, limit, presences, userIDs, keep):
		if self.gatewayobj.READY:
			saveAsQuery = query.lower() if saveAsQueryOverride==None else saveAsQueryOverride.lower()
			#create a spot to put the data in bot.gateway.guildMemberSearches
			if userIDs: #userID storage
				for i in guildIDs:
					if i not in self.gatewayobj.guildMemberSearches:
						self.gatewayobj.guildMemberSearches[i] = {"ids":set()}
					if "ids" not in self.gatewayobj.guildMemberSearches[i]:
						self.gatewayobj.guildMemberSearches[i]["ids"] = set()
			else: #query storage (saveAsQuery)
				for k in guildIDs:
					if k not in self.gatewayobj.guildMemberSearches:
						self.gatewayobj.guildMemberSearches[k] = {"queries":{}}
					if "queries" not in self.gatewayobj.guildMemberSearches[k]:
						self.gatewayobj.guildMemberSearches[k]["queries"] = {}
					if saveAsQuery not in self.gatewayobj.guildMemberSearches[k]["queries"]:
						self.gatewayobj.guildMemberSearches[k]["queries"][saveAsQuery] = set()
			self.gatewayobj.command(
				{
					"function": self.handleGuildMemberSearches,
					"priority": 0,
					"params": {
						"guildIDs": guildIDs,
						"saveAsQuery": saveAsQuery,
						"isQueryOverridden": saveAsQueryOverride != None,
						"userIDs": userIDs,
						"keep": keep,
					},
				}
			)
			self.gatewayobj.request.searchGuildMembers(guildIDs, query, limit, presences, userIDs)



#request

class UserRequest(object):
	__slots__ = ['gatewayobj']
	def __init__(self, gatewayobj):
		self.gatewayobj = gatewayobj

	def setStatus(self, status, activities, afk, since): #note, custom status goes in activities
		data = {
		    "op": self.gatewayobj.OPCODE.PRESENCE_UPDATE,
		    "d": {
		        "status": status,
		        "since": since,
		        "activities": activities,
		        "afk": afk
		    }
		}
		self.gatewayobj.send(data)


import asyncio
import time

from playwright.async_api import Error as PlaywrightError, TimeoutError as PlaywrightTimeoutError
from playwright.async_api import async_playwright, Browser, ProxySettings, Page, Playwright
from playwright_stealth import stealth_async
from yarl import URL


def proxy_url_to_playwright_proxy(proxy: str) -> ProxySettings:
    proxy = URL(proxy)
    return ProxySettings(
        server=f"{proxy.scheme}://{proxy.host}:{proxy.port}",
        password=proxy.password,
        username=proxy.user,
    )


class BaseBrowser:
    def __init__(
            self,
            *,
            headless=False,
            default_proxy: str = None,
    ):
        self._playwright: Playwright | None = None
        self._browser: Browser | None = None
        self._headless = headless
        self._default_proxy = proxy_url_to_playwright_proxy(default_proxy) if default_proxy else None

    async def create_browser(self):
        self._playwright = await async_playwright().start()
        self._browser = await self._playwright.chromium.launch(
            headless=self._headless, proxy=self._default_proxy)

    async def close_browser(self):
        await self._browser.close()
        await self._playwright.stop()

    async def __aenter__(self):
        await self.create_browser()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_browser()



from ..types import Types

#parse (remember to do static methods, unless you're changing the formatting)
class StartParse(object): #really hope this doesn't take too long to run...
	@staticmethod
	def ready(response):
		ready_data = dict(response["d"])
		ready_data.pop("merged_members")
		user_pool = {h["id"]:h for h in response["d"]["users"]} #convert to dict for faster retrieval
		#parse relationships
		ready_data["relationships"] = {i["id"]:dict(dict(i, **{"type":Types.RELATIONSHIP[i["type"]]}), **user_pool.get(i["id"], {})) for i in response["d"]["relationships"]}
		#parse private channels
		ready_data["private_channels"] = {}
		for j in response["d"]["private_channels"]:
			ready_data["private_channels"][j["id"]] = dict(j, **{"type":Types.CHANNEL[j["type"]]})
			if "recipient_ids" in ready_data["private_channels"][j["id"]]:
				recipient_ids = ready_data["private_channels"][j["id"]].pop("recipient_ids")
				ready_data["private_channels"][j["id"]]["recipients"] = {q:user_pool.get(q,{}) for q in recipient_ids}
		#add activities key to user settings
		ready_data["user_settings"]["activities"] = {}
		#parse guilds
		guilds = response["d"]["guilds"]
		ready_data["guilds"] = {k["id"]:k for k in guilds}
		for personal_role, guild in zip(response["d"]["merged_members"], guilds):
			if "unavailable" not in ready_data["guilds"][guild["id"]]:
				#take care of emojis
				if isinstance(guild["emojis"], list):
					ready_data["guilds"][guild["id"]]["emojis"] = {l["id"]:l for l in guild["emojis"]}
				#take care of roles
				if isinstance(guild["roles"], list):
					ready_data["guilds"][guild["id"]]["roles"] = {m["id"]:m for m in guild["roles"]}
				#take care of channels
				if isinstance(guild["channels"], list):
					ready_data["guilds"][guild["id"]]["channels"] = {n["id"]:dict(n, **{"type":Types.CHANNEL[n["type"]]}) for n in guild["channels"]}
			#take care of personal role/nick
			ready_data["guilds"][guild["id"]]["my_data"] = next((i for i in personal_role if i["user_id"]==response["d"]["user"]["id"]), {}) #personal_role
			#take care of members
			ready_data["guilds"][guild["id"]]["members"] = {}
		return ready_data

	@staticmethod
	def ready_supplemental(response):
		ready_supp_data = dict(response["d"])
		ready_supp_data["online_friends"] = {o["user_id"]:o for o in response["d"]["merged_presences"]["friends"]}
		ready_supp_data.pop("guilds")
		ready_supp_data["voice_states"] = {p["id"]:p.get("voice_states",[]) for p in response["d"]["guilds"]} #id is the guild_id
		return ready_supp_data
